/**
  *  \file filenames.cpp
  */

#include <windows.h>
#include <cstdlib>
#include <fstream>
#include <ctype.h>
#include "filenames.h"

/** Extract path of a file name. */
std::string
getDirectoryName(std::string s)
{
    std::string::size_type n = s.find_last_of("/\\:");
    if (n == s.npos)
        return std::string("");
    return s.substr(0, n);
}

std::string
getBaseName(std::string s)
{
    std::string::size_type n = s.find_last_of("/\\:");
    if (n == s.npos)
        return s;
    return s.substr(n);
}

static void
splitDrive(std::string& dir, std::string& drive)
{
    if (dir.size() >= 2 && dir[1] == ':') {
        drive.assign(dir, 0, 2);
        drive[0] = toupper((unsigned char) drive[0]);
        dir.erase(0, 2);
    }
}

static bool
isSlash(char c)
{
    return c == '/' || c == '\\';
}

/** Given a parsed file:// url, get the actual file name.
    \param host u.getHost()
    \param s    u.getPath()
    \return file name

    \todo if we support UNC paths, we could convert
    file://host/path into \\host\path */
std::string
getFileNameFromUrl(std::string host, std::string s)
{
    (void) host;

    // flip slashes
    for (std::string::size_type i = 0; i < s.size(); ++i)
        if (s[i] == '/')
            s[i] = '\\';

    if (s.size() && s[0] == '\\')
        s.erase(0, 1);
    
    return s;
}

/** Convert file name into file:// url. The identity should be that
    getFileUrl(getFileNameFromUrl(...)) yields a file name identifying
    the same file.

    \todo if we support UNC paths, we could convert \\host\path into
    file://host/path, not file://///host/path as we do now */
std::string
getFileUrl(std::string name)
{
    // flip slashes
    for (std::string::size_type i = 0; i < name.size(); ++i)
        if (name[i] == '\\')
            name[i] = '/';
    return "file:///" + name;
}

/** Make file name. Returns the file name that will be accessed when
    we open \c name while sitting in working directory \c basedir. */
std::string
makeFileName(std::string basedir, std::string name)
{
    if (name.size() == 0)
        return basedir;

    if (basedir.size() == 0)
        return name;

    std::string basedrive, namedrive;
    splitDrive(basedir, basedrive);
    splitDrive(name, namedrive);

    /* In case name doesn't have a drive, copy it */
    if (namedrive.size() == 0)
        namedrive = basedrive;

    if (name.size() && isSlash(name[0])) {
        /* x:\foo, \foo */
        return namedrive + name;
    }

    if (namedrive == basedrive || namedrive.size() == 0) {
        /* x:foo, foo */
        if (basedir.empty() || !isSlash(basedir[basedir.size()-1]))
            basedir.append("\\");
        return basedrive + basedir + name;
    }

    /* we're here if both specified different drives, but name is not
       absolute. In this case, make it absolute. */
    return namedrive + "\\" + name;
}

std::string
getSettingsFileName()
{
    static const char FILENAME[] = ".vgapintegrator";
    const char* p = std::getenv("APPDATA");
    if (p && *p) {
        return makeFileName(p, FILENAME);
    } else {
        char buf[MAX_PATH];
        GetWindowsDirectory(buf, sizeof(buf));
        if (*buf)
            return makeFileName(buf, FILENAME);
        else
            return FILENAME;
    }
}

bool
isExistingFileName(std::string name)
{
    // no wildcards
    if (name.find_first_of("*?") != name.npos)
        return false;

    // check it
    WIN32_FIND_DATA d;
    HANDLE h = FindFirstFile(name.c_str(), &d);
    if (h != INVALID_HANDLE_VALUE) {
        FindClose(h);
        return true;
    } else {
        return false;
    }
}

bool
createDirectories(std::string name)
{
    // Create it?
    if (CreateDirectory(name.c_str(), 0))
        return true;

    // Why didn't we create?
    DWORD err = GetLastError();
    if (err == ERROR_ALREADY_EXISTS || err == ERROR_FILE_EXISTS) {
        // it already exists
        if (isExistingFileName(name))
            return true;
    }

    if (err == ERROR_PATH_NOT_FOUND || err == ERROR_FILE_NOT_FOUND) {
        // a path component does not exist
        std::string::size_type i = name.size();
        while (i && !isSlash(name[i-1]) && name[i-1] != ':')
            --i;
        while (i && isSlash(name[i-1]))
            --i;
        if (i && name[i-1] == ':')
            return false;
        createDirectories(std::string(name, 0, i));
        return CreateDirectory(name.c_str(), 0) != 0;
    }

    return false;
}

#ifdef MKDIRTEST
#include <iostream>
int main(int, char** argv)
{
    while (*++argv) {
        std::cout << "mkdir(" << *argv << ") = "
                  << createDirectories(*argv) << "\n";
    }
    return 0;
}
#endif
