/**
  *  \file sem.h
  *
  *  This file is called "sem.h" to avoid ambiguities with
  *  <semaphore.h>.
  */
#ifndef SEM_H_INCLUDED
#define SEM_H_INCLUDED

#include <windows.h>

/** Counting semaphore. We do not need nestable mutexes, so simple
    semaphores are enough. */
class Semaphore {
    HANDLE m_TheSemaphore;

    Semaphore(const Semaphore&);
    Semaphore& operator=(const Semaphore&);
 public:
    /** Initialize semaphore.
        \param value initial value, i.e. number of times wait() can
        be called without actually waiting. */
    explicit Semaphore(int value)
        { m_TheSemaphore = CreateSemaphore(0, value, 0x7FFFFFFF, 0); }

    ~Semaphore()
        { CloseHandle(m_TheSemaphore); }

    /** Release semaphore. Increments the semaphore. */
    void release()
        { ReleaseSemaphore(m_TheSemaphore, 1, 0); }

    /** Obtain semaphore. If the semaphore is zero, waits until
        someone calls release(), otherwise just decrements the
        semaphore. */
    void wait()
        { WaitForSingleObject(m_TheSemaphore, INFINITE); }

    /** Try to obtain the semaphore. If the semaphore is zero, returns
        false. Otherwise, decrements it and returns true. */
    bool tryWait()
        { return WaitForSingleObject(m_TheSemaphore, 0) == WAIT_OBJECT_0; }
};

#endif
