/**
  *  \file initializer.cpp
  *  \brief Game Initialisation Tasks
  */

#include <vector>
#include "initializer.h"
#include "data.h"
#include "game.h"
#include "basicrequests.h"
#include "filenames.h"

/** Create initializer. */
Initializer::Initializer(Data& data)
    : m_Data(data),
      m_HaveMissingSpecs(0),
      m_HaveMissingDirs(0)
{ }

/** Assign game directories to games that don't have one.
    \param force Force assignment for games outside the Winplan tree */
void
Initializer::assignGameDirectories(bool force)
{
    class GameDirProvider : public ValueEnumerator<Game> {
        Data& d;
        bool force;
     public:
        GameDirProvider(Data& d, bool force)
            : d(d), force(force)
            { }
        void handleItem(Game& g)
            {
                if (g.getStatus() == Game::StatusNew || g.getLocalDirectory().empty()) {
                    g.setLocalDirectory(std::string());
                    std::string n = d.getNewGameDirectory(g.getName(), force);
                    if (!n.empty()) {
                        g.setLocalDirectory(n);
                        g.setStatus(Game::StatusNeedSpecs);
                    } else {
                        g.setStatus(Game::StatusNew);
                    }
                }

                if (g.getStatus() != Game::StatusNew) {
                    // create directory if needed
                    if (!isExistingFileName(g.getLocalDirectory())) {
                        // FIXME: do we want error handling here?
                        createDirectories(g.getLocalDirectory());
                        g.setStatus(Game::StatusNeedSpecs);
                    }
                    // FIXME: create host directories here?
                }
            }
    };

    GameDirProvider gdp(m_Data, force);
    m_Data.enumerateGames(gdp);

    check();
}

/** Download missing specification files. Posts a request to download
    the specifications for all games marked as new. */
void
Initializer::downloadMissingSpecs(RequestConsumer& con)
{
    class SpecDownloader : public ValueEnumerator<Game> {
        RequestConsumer& con;
     public:
        SpecDownloader(RequestConsumer& con)
            : con(con)
            { }
        void handleItem(Game& g)
            {
                if (g.getStatus() == Game::StatusNeedSpecs) {
                    doDownloadSpec(g, con);
                    g.setStatus(Game::StatusOK); /* FIXME!!! */
                }
            }
    };

    SpecDownloader sd(con);
    m_Data.enumerateGames(sd);

    check();
}

/** Re-determine agenda. Checks whether there are any games that need
    directory assignment or spec download. */
void
Initializer::check()
{
    m_HaveMissingSpecs = 0;
    m_HaveMissingDirs = 0;

    // figure out games
    std::vector<Game*> games;
    VectorValueEnumerator<Game> game_enum(games);
    m_Data.enumerateGames(game_enum);

    // do it
    for (std::vector<Game*>::iterator i = games.begin(); i != games.end(); ++i) {
        Game& g = **i;
        if (g.getStatus() == Game::StatusNew || g.getLocalDirectory().empty())
            ++m_HaveMissingDirs;
        else if (g.getStatus() == Game::StatusNeedSpecs)
            ++m_HaveMissingSpecs;
    }
}
