/**
  *  \file net/connection.cpp
  *  \brief Network Connection
  *  \author Stefan Reuther
  */

#include <errno.h>
#include "net/connection.h"
#include "nls.h"

/** Create blank connection object. */
Connection::Connection()
    : m_Socket()
{ }

/** Destructor. Closes the socket in case it is still open. */
Connection::~Connection()
{
    close();
}

/** Connect. Looks up the host in the host database, opens a socket
    and connects. In case this Connection object already holds an open
    connection, closes that first.

    \param host   Name of host to connect to
    \param port   Port number

    \return true on success, false on error (+message set) */
bool
Connection::connect(std::string host, int port)
{
    // close if we are already connected
    close();

    // now resolve the name we want to look at
    struct hostent* hp = gethostbyname(host.c_str());
    if (hp == 0) {
        setMessage(_("Host \"") + host + _("\" not found."), Socket::getResolverError());
        return false;
    }

    struct sockaddr_in server_address;
    if (hp->h_length != sizeof(server_address.sin_addr)) {
        setMessage(_("Internal error"), _("Invalid h_length in return of gethostbyname"));
        return false;
    }
    memcpy(&server_address.sin_addr, hp->h_addr, sizeof(server_address.sin_addr));
    server_address.sin_family = hp->h_addrtype;
    server_address.sin_port = htons(port);

    // connect
    Socket s(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (!s.isValid()) {
        setMessage(_("Internal error"), s.getError());
        return false;
    }
    if (!s.connect((sockaddr*) &server_address, sizeof(server_address))) {
        setMessage(_("Unable to connect to \"") + host + "\"", s.getError());
        return false;
    }

    m_Socket = s;
    return true;
}

/** Disconnect. Null operation in case we are not connected. */
void
Connection::close()
{
    if (m_Socket.isValid())
        m_Socket.close();
}

/** Write data.
    \param data [in] Data to write
    \param size [in] Number of bytes in data.
    \return true on success, false on error (+message set) */
bool
Connection::write(const void* data, std::size_t size)
{
    int rv = m_Socket.write(data, size);
    if (rv == int(size)) {
        return true;
    } else {
        setMessage(_("Write error"), m_Socket.getError());
        return false;
    }
}

/** Write string.
    \param data [in] Data to write
    \return true on success, false on error (+message set) */
bool
Connection::write(const std::string& data)
{
    return write(data.data(), data.size());
}

/** Read single line from input. Accepts Network and Unix format.
    \param result [in/out] Line. Note that received characters are appended.
    \return true if complete line read, false otherwise (+message set) */
bool
Connection::readLine(std::string& result)
{
    char c;
    while (m_Socket.read(&c, 1) == 1) {
        if (c == '\r')
            /* ignore */;
        else if (c == '\n')
            return true;
        else
            result += c;
    }
    setMessage(_("Read error"), _("Unable to receive reply"));
    return false;
}

/** Set error message.
    \param ourmsg User-friendly error message.
    \param srvmsg Server or system message ("404 Not Found", "No route to host") */
void
Connection::setMessage(std::string ourmsg, std::string srvmsg)
{
    m_OurMessage = ourmsg;
    m_ServerMessage = srvmsg;
}
