/**
  *  \file program.cpp
  *  \brief Program class implementation
  *  \author Stefan Reuther
  */

#include "program.h"
#include "action.h"
#include "util.h"
#include "nls.h"
#include "filenames.h"
#include "loaderstate.h"

/** Constructor.
    \param name Program name (user-visible). */
Program::Program(std::string name)
    : m_Name(name), m_Preference(0)
{ }

/** Destructor. */
Program::~Program()
{
    safeClear(m_Actions);
}

bool
Program::assign(std::string name, std::string value, LoaderState& st)
{
    if (name == "-") {
        m_Name = value;
        return true;
    }
    if (name == "directory") {
        m_Directory = st.expandDirectory(value);
        return true;
    }
    if (name == "version") {
        m_Version = value;
        return true;
    }
    if (name == "description") {
        m_Description = value;
        return true;
    }
    if (name == "supports") {
        m_SupportedFeatures.assignFromString(value);
        return true;
    }
    if (name == "provides") {
        m_ProvidedFeatures.assignFromString(value);
        return true;
    }
    if (name == "preference") {
        return atoi(value, m_Preference);
    }
    return false;
}

void
Program::enumerate(KVEnumerator& e)
{
    e.handleItem("-",            _("Name of the program"), m_Name);
    e.handleItem("directory",    _("Directory"), m_Directory);
    e.handleItem("version",      _("Version of the program"), m_Version);
    e.handleItem("description",  _("More detailed description"), m_Description);
    e.handleItem("supports",     _("List of supported features"), m_SupportedFeatures.toString());
    e.handleItem("provides",     _("List of provided features"), m_ProvidedFeatures.toString());
    e.handleItem("preference",   _("Preference rating, bigger is better"), itoa(m_Preference));
}

PropertyObject*
Program::createSubtree(std::string name, std::string value, LoaderState& /*st*/)
{
    if (name == "action") {
        return getCreateActionByName(value);
    }
    return 0;
}

/** Get name of this program. */
std::string
Program::getName() const
{
    return m_Name;
}

/** Get directory of this program. This is the "directory" property. */
std::string
Program::getDirectory() const
{
    return m_Directory;
}

/** Set directory of this program.
    \param s directory, should be absolute. */
void
Program::setDirectory(std::string s)
{
    m_Directory = s;
}

/** Check for feature. Checks the "provides" property.
    \param feature Feature to check for
    \return true iff we support that feature */
bool
Program::hasFeature(std::string feature)
{
    return m_ProvidedFeatures.contains(feature);
}

/** Get preference metric. This is a quality measure; the higher this
    value, the better this program. */
int
Program::getPreference() const
{
    return m_Preference;
}

/** Add an action.
    \param a Action to add. Must have been created with a reference to
             this Program */
void
Program::addAction(Action* a)
{
    safeAppend(m_Actions, a);
}

/** Find action by name.
    \param s Action name to search for
    \return pointer to action, 0 if none found */
Action*
Program::getActionByName(std::string s)
{
    for (std::vector<Action*>::iterator i = m_Actions.begin(); i != m_Actions.end(); ++i)
        if ((**i).getName() == s)
            return *i;
    return 0;
}

/** Find action by name, create if it does not exist.
    \return action */
Action*
Program::getCreateActionByName(std::string s)
{
    Action* a = getActionByName(s);
    if (!a) {
        a = new Action(s, *this);
        addAction(a);
    }
    return a;
}

/** Enumerate this program's actions. */
void
Program::enumerateActions(ValueEnumerator<Action>& e)
{
    for (std::vector<Action*>::iterator i = m_Actions.begin(); i != m_Actions.end(); ++i)
        e.handleItem(**i);
}
