/**
  *  \file selection.cpp
  *  \brief Selection class implementation
  *  \author Stefan Reuther
  */

#include "selection.h"
#include "game.h"
#include "race.h"
#include "action.h"
#include "program.h"
#include "expander.h"
#include "util.h"

/** Constructor. Initially, nothing is selected. */
Selection::Selection()
    : m_Game(0), m_Race(0), m_Action(0)
{ }

/** Destructor. */
Selection::~Selection()
{ }

/** Select a game. Automatically selects the default race for
    that game if possible.
    \param g Game to select. Can be 0 for none. */
void
Selection::selectGame(Game* g)
{
    if (m_Game != g)
        m_Race = 0;
    m_Game = g;
    recheckGame();
}

/** Select a race.
    \param r Race to select. Can be 0 for none. Must belong to the
             selected game if any. */
void
Selection::selectRace(Race* r)
{
    m_Race = r;
}

/** Select an action.
    \param a Action to select. Can be 0 for none. */
void
Selection::selectAction(Action* a)
{
    m_Action = a;
}

/** Recheck game. Call this after every change to the game's race
    set. */
void
Selection::recheckGame()
{
    if (m_Game) {
        if (!m_Game->getDefaultRace())
            m_Race = 0;
        else if (!m_Race)
            m_Race = m_Game->getDefaultRace();
    }
}

/** Set all variables on the expander. */
void
Selection::setVariables(Expander& exp) const
{
    // Game
    if (m_Game != 0) {
        exp.addObject(*m_Game);
        exp.addVariable("game", m_Game->getName());
    }

    // Race
    if (m_Race != 0) {
        exp.addObject(*m_Race);
        exp.addVariable("race", itoa(m_Race->getRaceNumber()));
    }

    // Action / Program
    if (m_Action != 0) {
        exp.addObject(m_Action->getProgram());
        exp.addVariable("program", m_Action->getProgram().getName());
        exp.addObject(*m_Action);
        exp.addVariable("action", m_Action->getName());
    }
}

/** Expand URL/command line string. Expands the string with all
    currently known variables set.
    \return expanded string. Empty string if it refers to a variable
    which is currently unknown. */
std::string
Selection::expandString(std::string input) const
{
    Expander x;
    setVariables(x);
    return x.expand(input);
}

/** Compare two selections for equality. */
bool
Selection::operator==(const Selection& other) const
{
    return m_Game == other.m_Game
        && m_Race == other.m_Race
        && m_Action == other.m_Action;
}
