/**
  *  \file transaction.cpp
  *  \brief Transactions based on Multiple Reader/Writer Locks
  */

#include <assert.h>
#include "transaction.h"

/** Global mutex. All operations that manipulate lock counters or
    s_WaitingTransactions must hold this semaphore. */
Semaphore Transaction::s_GlobalLock(1);

/** List of all waiting transactions. All transactions in this list
    want to wake up when they can take their locks, and are hanging in
    their m_WaitLock.wait(). */
Transaction* Transaction::s_WaitingTransactions(0);

/** Create new blank transaction. */
Transaction::Transaction()
    : m_Active(false),
      m_WaitLock(0),
      m_Next(0),
      m_Prev(0)
{ }

/** Destroy transaction. If it is still active, releases all its locks
    first. */
Transaction::~Transaction()
{
    releaseLocks();
}

/** Add a read operation. */
void
Transaction::addReadOperation(Transactable* obj)
{
    assert(!m_Active);
    m_Operations.push_back(Element(obj, ReadOperation));
}

/** Add a write operation. */
void
Transaction::addWriteOperation(Transactable* obj)
{
    assert(!m_Active);
    m_Operations.push_back(Element(obj, WriteOperation));
}

/** Acquires all read and write locks needed to start this
    transactions. Returns when all the locks are held. For objects
    added with addReadOperation(), this means the objects have no
    writer. For objects added with addWriteOperation(), this means we
    are the only writer. */
void
Transaction::acquireLocks()
{
    if (!m_Active) {
        s_GlobalLock.wait();
        if (checkCanStart()) {
            doStart();
            s_GlobalLock.release();
        } else {
            m_Next = s_WaitingTransactions;
            m_Prev = &s_WaitingTransactions;
            s_GlobalLock.release();
            m_WaitLock.wait();
        }
    }
}

/** Try to acquire all locks. Same as acquireLocks(), but does not
    wait if locks cannot immediately be taken.
    \return true if locks are held, false if some are not available */
bool
Transaction::tryAcquireLocks()
{
    if (!m_Active) {
        s_GlobalLock.wait();
        if (checkCanStart()) {
            doStart();
            s_GlobalLock.release();
            return true;
        } else {
            s_GlobalLock.release();
            return false;
        }
    } else {
        return true;
    }
}

/** Release all locks. Wakes up pending other tasks that may be
    runnable now. */
void
Transaction::releaseLocks()
{
    if (m_Active) {
        s_GlobalLock.wait();
        /* First, release our locks */
        doStop();

        /* Now wake up everyone */
        Transaction* next;
        for (Transaction* p = s_WaitingTransactions; p != 0; p = next) {
            next = p->m_Next;
            if (p->checkCanStart()) {
                p->doStart();
                *p->m_Prev = p->m_Next;
                p->m_Prev = 0;
                p->m_Next = 0;
                p->m_WaitLock.release();
            }
        }
        s_GlobalLock.release();
    }
}

/** Check whether all locks can be taken.
    \pre s_GlobalLock held */
bool
Transaction::checkCanStart()
{
    for (std::vector<Element>::iterator i = m_Operations.begin(); i != m_Operations.end(); ++i) {
        if (i->m_Operation == WriteOperation) {
            if (i->m_Object->m_NumReaders != 0 || i->m_Object->m_NumWriters != 0)
                return false;
        } else {
            if (i->m_Object->m_NumWriters != 0)
                return false;
        }
    }
    return true;
}

/** Take all locks.
    \pre s_GlobalLock held */
void
Transaction::doStart()
{
    for (std::vector<Element>::iterator i = m_Operations.begin(); i != m_Operations.end(); ++i) {
        if (i->m_Operation == WriteOperation)
            i->m_Object->m_NumWriters++;
        else
            i->m_Object->m_NumReaders++;
    }
    m_Active = true;
}

/** Release all locks.
    \pre s_GlobalLock held. */
void
Transaction::doStop()
{
    for (std::vector<Element>::iterator i = m_Operations.begin(); i != m_Operations.end(); ++i) {
        if (i->m_Operation == WriteOperation)
            i->m_Object->m_NumWriters--;
        else
            i->m_Object->m_NumReaders--;
    }
    m_Active = false;
}
