/**
  *  \file turnuploadparser.cpp
  *  \brief TurnUploadParser class
  *  \author Stefan Reuther
  */

#include <cstring>
#include <cctype>
#include "turnuploadparser.h"
#include "util.h"
#include "nls.h"

/** Find substring, case-blind.
    \param data         [in] Haystack to search in
    \param text         [in] Needle to search for
    \param start_index  [in] First position in data to look at
    \return index into data, such that return >= start_index, and
    data.substr(return, strlen(text)) == text (case-blind). npos if no
    such position exists. */
static std::string::size_type
strFindCaseBlind(const std::string& data, const char* text, std::string::size_type start_index = 0)
{
    using namespace std;
    size_t len = strlen(text);
    if (len > data.size())
        return std::string::npos;

    size_t max = data.size() - len;
    for (size_t i = start_index; i < max; ++i) {
        for (size_t x = 0; x < len; ++x) {
            if (toupper((unsigned char) data[i + x]) != toupper((unsigned char) text[x]))
                goto next;
        }
        return i;
     next:;
    }
    return std::string::npos;
}

/** Extract delimited text. Search is case-insensitive.
    \param data     [in] Haystack to search in
    \param begin    [in] Beginning delimiter
    \param end      [in] End delimiter
    \return Text appearing between first occurrence of begin and
    end in data. */
static std::string
extractText(const std::string& data, const char* begin, const char* end)
{
    using namespace std;
    std::string::size_type b = strFindCaseBlind(data, begin);
    if (b == data.npos)
        return std::string();

    b += strlen(begin);
    std::string::size_type e = strFindCaseBlind(data, end, b);

    if (e == data.npos)
        return std::string(data, b);
    else
        return std::string(data, b, e - b);
}

/** Extract keyword. Given an arbitrary string, extract that which
    looks like a keyword. */
static std::string
extractKeyword(std::string text)
{
    using namespace std;
    std::string::size_type i = 0;
    while (i < text.size() && isalnum((unsigned char) text[i]))
        ++i;
    return text.substr(0, i);
}

/** Remove HTML tags and normalize text. This strips all tags and
    their attributes and removes successive blank lines. The result
    should be something vaguely readable even in case the original was
    weird HTML. The idea is that the stuff we're interested in is just
    chktrn / PHost output wrapped in a \<pre\> anyway. */
static std::string
removeTags(std::string inp)
{
    using namespace std;
    // State of parser
    enum { sIdle, sInEntity, sInTag, sInQuote } state = sIdle;
    // State of output
    bool previous_line_was_blank = true;
    bool this_line_was_nonblank = false;
    // Output accumulator
    std::string output;
    // Line accumulator
    std::string accum;
    // Entity
    std::string entity;

    for (std::string::size_type x = 0; x < inp.size(); ++x) {
        char c = inp[x];
        switch (state) {
         case sIdle:
            if (c == '<') {
                state = sInTag;
            } else if (c == '&') {
                state = sInEntity, entity.clear();
            } else if (c == '\r') {
                ;
            } else if (c == '\n') {
                if (this_line_was_nonblank) {
                    output += accum;
                    output += '\n';
                    previous_line_was_blank = false;
                    this_line_was_nonblank = false;
                } else {
                    if (!previous_line_was_blank)
                        output += '\n';
                    previous_line_was_blank = true;
                    this_line_was_nonblank = false;
                }
                accum.clear();
            } else {
                accum += c;
                if (!isspace((unsigned char) c))
                    this_line_was_nonblank = true;
            }
            break;
         case sInEntity:
            if (c == '\r') {
                ;
            } else if (c == ';' || c == '\n') {
                if (entity == "lt")
                    accum += '<', this_line_was_nonblank = true;
                else if (entity == "gt")
                    accum += '>', this_line_was_nonblank = true;
                else if (entity == "amp")
                    accum += '&', this_line_was_nonblank = true;
                else if (entity == "quot")
                    accum += '"', this_line_was_nonblank = true;
                else
                    /* FIXME */;
                state = sIdle;
            } else {
                entity += c;
            }
            break;
         case sInTag:
            if (c == '"')
                state = sInQuote;
            else if (c == '>')
                state = sIdle;
            break;
         case sInQuote:
            if (c == '"')
                state = sInTag;
            break;
        }
    }

    // last line
    if (accum.size() && this_line_was_nonblank) {
        output += accum;
        output += '\n';
    }

    return output;
}

/** Constructor. */
TurnUploadParser::TurnUploadParser()
{
    setUnknown();
}

/** Destructor. */
TurnUploadParser::~TurnUploadParser()
{ }

/** Parse string. Fills in our fields accordingly. Since any input is
    valid (but possibly information-free), this has no failure mode. */
void
TurnUploadParser::parse(std::string s)
{
    // Checkturn result
    m_CheckturnResult = extractText(s, "<!--ATUP-Begin-Checkturn-->", "<!--ATUP-End-Checkturn-->");
    if (m_CheckturnResult.empty()) {
        m_CheckturnResult = extractText(s, "<body", "</body");
        /* result starts with a '>' from the body tag. Add back the
           \<body, so removeTags() will remove it correctly, including
           its attributes. */
        if (m_CheckturnResult.size())
            m_CheckturnResult.insert(0, "<body");
    }
    /* remove tags */
    m_CheckturnResult = removeTags(m_CheckturnResult);

    // Registration
    m_Registration = strLower(extractKeyword(strTrim(extractText(s, "<!--ATUP-Registration:", "-->"))));
    if (m_Registration.empty())
        m_Registration = "unknown";

    // Result
    m_Result = strLower(extractKeyword(strTrim(extractText(s, "<!--ATUP-Result:", "-->"))));
    if (m_Result.empty()) {
        m_Result = "unknown";
        /* Attempt to parse checkturn output. Note that removeTags has
           already removed \r's, so we can search for text containing
           \n's. */
        std::string tmp = extractText(m_CheckturnResult, "PHost is returning error code", "\n");
        int i;
        if (tmp.size() && strFindCaseBlind(m_CheckturnResult, "blank -- Player does not exist") == std::string::npos) {
            /* This is PHost, saying that the turn is bad. Note that
               if this is a "phost -c0" log, we cannot say whether
               it's our turn that was bad. Okay, we could parse the
               table, but that's too much hassle for now.  */
            if (atoi(tmp, i)) {
                if (i & 128)
                    m_Result = "red";
                else if (i & 64)
                    m_Result = "yellow";
                else if (i & (4+8+16+32))
                    m_Result = "format";
                else if (i & 2)
                    m_Result = "stale";
                else if (i & 1)
                    m_Result = "bad";   // means: turn not found
            }
        } else if (strFindCaseBlind(m_CheckturnResult, "\nNo errors found.\n") != std::string::npos) {
            /* This is PHost, saying that the turn is ok */
            /* FIXME: to avoid false security, it would probably be
               wise to avoid declaring the turn good here. Maybe the
               host has other reasons to declare it bad. */
            m_Result = "good";
        }
    }
}

/** Set status to "unknown". */
void
TurnUploadParser::setUnknown()
{
    m_CheckturnResult.clear();
    m_Result       = "unknown";
    m_Registration = "unknown";
}

/** Set status to "mailed". */
void
TurnUploadParser::setMailed()
{
    m_CheckturnResult.clear();
    m_Result       = "mailed";
    m_Registration = "mailed";
}

/** Set status to "copied". */
void
TurnUploadParser::setCopied()
{
    m_CheckturnResult.clear();
    m_Result       = "copied";
    m_Registration = "copied";
}

namespace {
    struct Info {
        int result;
        const char* keyword;
        const char* text;
    };
};

static const Info result_info[] = {
    { TurnUploadParser::ResultGood,    "good",    N_("Turn was accepted.") },
    { TurnUploadParser::ResultBad,     "bad",     N_("Turn was not accepted.") },
    { TurnUploadParser::ResultRed,     "red",     N_("Turn was rejected with a RED error.") },
    { TurnUploadParser::ResultYellow,  "yellow",  N_("Turn was rejected with a YELLOW error.") },
    { TurnUploadParser::ResultBad,     "stale",   N_("Turn was from wrong turn (stale).") },
    { TurnUploadParser::ResultBad,     "format",  N_("Turn file was damaged.") },
    { TurnUploadParser::ResultBad,     "reg",     N_("Registration key had problems.") },
    { TurnUploadParser::ResultUnknown, "unknown", N_("Turn status not known.") },
    { TurnUploadParser::ResultUnknown, "mailed",  N_("Turn mailed, status not known.") },
    { TurnUploadParser::ResultUnknown, "copied",  N_("Turn copied, status not known.") },
};

static const Info reg_info[] = {
    { TurnUploadParser::RegGood,       "registered", N_("Registered turn recognized.") },
    { TurnUploadParser::RegShareware,  "shareware",  N_("Shareware turn recognized.") },
    { TurnUploadParser::RegBad,        "invalid",    N_("Registration key had problems.") },
    { TurnUploadParser::RegUnknown,    "unknown",    N_("Registration status not known.") },
    { TurnUploadParser::RegUnknown,    "mailed",     N_("Turn mailed, registration status not known.") },
    { TurnUploadParser::RegUnknown,    "copied",     N_("Turn copied, registration status not known.") },
};

/** Explain a result.
    \param keyword [in] Keyword received from server, see getResult()
    \param result  [out] Result class
    \return human-readable result */
std::string
TurnUploadParser::explainResult(std::string keyword, TurnResult& result)
{
    for (unsigned int i = 0; i < countof(result_info); ++i) {
        if (result_info[i].keyword == keyword) {
            result = TurnResult(result_info[i].result);
            return _(result_info[i].text);
        }
    }
    result = ResultUnknown;
    return _("Unknown answer from server (\"") + keyword + "\")";
}

/** Explain a registration status.
    \param keyword [in] Keyword received from server, see getRegistration()
    \param result  [out] Result class
    \return human-readable result */
std::string
TurnUploadParser::explainRegistration(std::string keyword, TurnRegistration& result)
{
    for (unsigned int i = 0; i < countof(reg_info); ++i) {
        if (reg_info[i].keyword == keyword) {
            result = TurnRegistration(reg_info[i].result);
            return _(reg_info[i].text);
        }
    }
    result = RegUnknown;
    return _("Unknown answer from server (\"") + keyword + "\")";
}
