/**
  *  \file workerthread.cpp
  */

#include "workerthread.h"
#include "scopedlock.h"
#include "util.h"

/** Create worker thread. Thread is created suspended, user must call
    start() to start it. */
WorkerThread::WorkerThread()
    : m_QueueMutex(1),
      m_PendingRequestSem(0),
      m_FinishedRequestsSem(0),
      m_PendingRequests(),
      m_FinishedRequests()
{
    setStackSize(32768);
}

WorkerThread::~WorkerThread()
{
    // FIXME...
}

/** Add new asynchronous request. The worker thread will obtain
    ownership of the request object, and immediately start performing
    it if it can. */
void
WorkerThread::pushRequest(Request* r)
{
    r->setStatus(Request::StatusWaiting);
    ScopedLock lock(m_QueueMutex);
    safeAppend(m_PendingRequests, r);
    m_PendingRequestSem.release();
}

/** Fetch a finished request. The returned object is hereby released
    from the worker thread's ownership, and user assumes
    responsibility. */
Request*
WorkerThread::popFinishedRequest()
{
    ScopedLock lock(m_QueueMutex);
    if (!m_FinishedRequests.empty()) {
        Request* p = m_FinishedRequests.front();
        m_FinishedRequests.pop_front();
        return p;
    } else {
        return 0;
    }
}

/** Worker thread function. */
void
WorkerThread::run()
{
    while (1) {
        Request* r;

        m_PendingRequestSem.wait();
        {
            ScopedLock lock(m_QueueMutex);
            r = m_PendingRequests.front();
            m_PendingRequests.pop_front();
        }
        r->setStatus(Request::StatusRunning);
        try {
            r->run();
            if (r->getStatus() == Request::StatusRunning)
                r->setStatus(Request::StatusSucceeded);
        }
        catch (...) {
            r->setStatus(Request::StatusFailed);
        }
        {
            // FIXME: if this safeAppend throws, the request
            // is lost, and the WorkerThread dies. I hope that
            // we don't run out of memory so often when allocating
            // a list node with one word payload.
            ScopedLock lock(m_QueueMutex);
            safeAppend(m_FinishedRequests, r);
        }
        m_FinishedRequestsSem.release();
    }
}

#ifdef WORKERTEST
#include <unistd.h>
#include <stdlib.h>
#include <iostream>
class TestRequest : public Request {
 public:
    int n, wtime;

    TestRequest(int n, int wtime)
        : n(n), wtime(wtime)
        { }
    void run()
        {
            std::cout << ">> Request " << n << " starts." << std::endl;
            sleep(wtime);
            std::cout << ">> Request " << n << " finished." << std::endl;
        }
    std::string getName()
        { return "TestRequest"; }
};

static void
logRequest(Request* r)
{
    if (TestRequest* tr = dynamic_cast<TestRequest*>(r)) {
        std::cout << tr->n << " with time " << tr->wtime << " finished" << std::endl;
    } else {
        std::cout << "unknown request finished" << std::endl;
    }
}

int main()
{
    int counter = 0;
    WorkerThread wt;
    wt.start();

    std::string line;
    while (std::cout << "?" << std::flush, getline(std::cin, line)) {
        if (line == "wait") {
            wt.getFinishSemaphore().wait();
            Request* r = wt.popFinishedRequest();
            if (!r)
                std::cout << "huh?" << std::endl;
            else
                logRequest(r);
        } else if (line.substr(0, 4) == "new ") {
            wt.pushRequest(new TestRequest(++counter, atoi(line.c_str()+4)));
        } else if (line == "q") {
            break;
        } else if (line == "") {
            while (Request* r = wt.popFinishedRequest()) {
                logRequest(r);
                wt.getFinishSemaphore().wait();
            }
        } else {
            std::cout << "unknown command" << std::endl;
        }
    }
}
#endif
