/**
  *  \file wingui/api.cpp
  *  \brief Windows API wrappers
  *
  *  This module provides a few lightweight Win32 API wrappers making
  *  the API calls a little easier to use from C++.
  */

#include "api.h"

/** Constructor. */
Window::Window()
{ }

/** Destructor. */
Window::~Window()
{ }

/** General event handler. This routine is called whenever Windows
    sends a message to a window associated with this instance of the
    Window class. */
LRESULT
Window::handle(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    return DefWindowProc(hwnd, msg, wparam, lparam);
}

LRESULT CALLBACK
Window::WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    if (msg == WM_CREATE) {
        // We could remember our pointer here
        LPCREATESTRUCT ps = (LPCREATESTRUCT) lparam;
        SetWindowLong(hwnd, GWL_USERDATA, LONG(ps->lpCreateParams));
        return 0;
    } else if (msg == WM_NCCREATE || msg == WM_NCDESTROY) {
        // Do not handle these, our pointer may already be dead
        return DefWindowProc(hwnd, msg, wparam, lparam);
    } else {
        Window* me = reinterpret_cast<Window*>(GetWindowLong(hwnd, GWL_USERDATA));
        if (me) {
            return me->handle(hwnd, msg, wparam, lparam);
        } else {
            return DefWindowProc(hwnd, msg, wparam, lparam);
        }
    }
}

/** Register a window class.
    \param name Name under which class is known in the Win32 API
    \param icon Icon */
void
Window::registerWindowClass(const char* name, HICON icon)
{
    WNDCLASS wc;
    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = Window::WindowProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = GetModuleHandle(0);
    wc.hIcon         = icon;
    wc.hCursor       = LoadCursor(0, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName  = 0;
    wc.lpszClassName = name;
    RegisterClass(&wc);
}

/** Create window. This is almost the same as the API's CreateWindow()
    function, but associates the window with this instance of Window
    resp. the class derived from Window. */
HWND
Window::create(LPCTSTR lpClassName,         ///< pointer to registered class name
               LPCTSTR lpWindowName,        ///< pointer to window name
               DWORD dwStyle,               ///< window style
               int x,                       ///< horizontal position of window
               int y,                       ///< vertical position of window
               int nWidth,                  ///< window width
               int nHeight,                 ///< window height
               HWND hWndParent,             ///< handle to parent or owner window
               HMENU hMenu)                 ///< handle to menu or child-window identifier
{
    return CreateWindow(lpClassName, lpWindowName, dwStyle,
                        x, y, nWidth, nHeight,
                        hWndParent, hMenu,
                        GetModuleHandle(0), this);
}

/** Get text of a window. Same as GetWindowText(), but returns a C++
    string. */
std::string
getWindowString(HWND wnd)
{
    int n = GetWindowTextLength(wnd);
    char* tmp = new char[n + 1];
    GetWindowText(wnd, tmp, n+1);
    tmp[n] = 0;
    std::string rv(tmp);
    delete[] tmp;
    return rv;
}

/** Get text of a dialog item. Same as GetDlgItemText(), but returns a
    C++ string. */
std::string
getDlgItemString(HWND dlg, int id)
{
    return getWindowString(GetDlgItem(dlg, id));
}
