/**
  *  \file wingui/statuspanel.cpp
  *  \brief Status Panel Widget
  *  \author Stefan Reuther
  */

#include "statuspanel.h"

static const char CLASS_NAME[] = "Wingui.Statuspanel.Class";

HFONT StatusPanel::m_BoldFont, StatusPanel::m_NormalFont;
int StatusPanel::m_DefaultWidth, StatusPanel::m_DefaultHeight;

/** Constructor.
    \param title   Title of the request. Displayed in black, boldface.
    \param status  Status of the request. Displayed in regular weight.
    \param statuscolor Color for \c status */
StatusPanel::StatusPanel(std::string title, std::string status, COLORREF statuscolor)
    : m_Title(title),
      m_Status(status),
      m_StatusColor(statuscolor),
      m_Hwnd(0),
      m_Age(0)
{
    if (!m_BoldFont && !m_NormalFont)
        init();
}

/** Destructor. Also destroys the Win32 window object. */
StatusPanel::~StatusPanel()
{
    if (m_Hwnd)
        DestroyWindow(m_Hwnd);
}

/** Draw this thing.
    \param hwnd    Our window handle
    \param dc      Device context (unused?) */
void
StatusPanel::draw(HWND hwnd, HDC dc)
{
    RECT rgn;
    if (!GetUpdateRect(hwnd, &rgn, FALSE))
        return;

    PAINTSTRUCT ps;
    dc = BeginPaint(hwnd, &ps);       // FIXME: do we need the 'dc' parameter?

    // Start by drawing five rectangles
    static const COLORREF rects[] = { RGB(255, 255, 255),
                                      RGB(0, 0, 0),
                                      RGB(255, 255, 255),
                                      RGB(0, 128, 255),
                                      RGB(0, 0, 255) };
    GetClientRect(hwnd, &rgn);
    for (int i = 0; i < 5; ++i) {
        HPEN p = CreatePen(PS_SOLID, 1, rects[i]);
        HGDIOBJ prev = SelectObject(dc, p);
        MoveToEx(dc, rgn.left, rgn.top, 0);
        LineTo(dc, rgn.right, rgn.top);
        LineTo(dc, rgn.right, rgn.bottom);
        LineTo(dc, rgn.left, rgn.bottom);
        LineTo(dc, rgn.left, rgn.top);
        SelectObject(dc, prev);
        DeleteObject(p);
        rgn.left++;
        rgn.top++;
        rgn.bottom--;
        rgn.right--;
    }

    // Now the interior
    FillRect(dc, &rgn, (HBRUSH) GetStockObject(WHITE_BRUSH));

    RECT textsize;
    textsize.left = rgn.left + 2;
    textsize.top = rgn.top;
    textsize.bottom = (rgn.top + rgn.bottom) / 2;
    textsize.right = rgn.right - 2;

    HGDIOBJ prev_font  = SelectObject(dc, m_BoldFont);
    SetTextColor(dc, RGB(0,0,0));
    DrawText(dc, m_Title.data(), m_Title.size(), &textsize, DT_END_ELLIPSIS | DT_LEFT | DT_VCENTER | DT_NOPREFIX);

    SetTextColor(dc, m_StatusColor);
    SelectObject(dc, m_NormalFont);
    textsize.left = rgn.left + 2;
    textsize.top = (rgn.top + rgn.bottom) / 2;
    textsize.bottom = rgn.bottom;
    textsize.right = rgn.right - 2;
    DrawText(dc, m_Status.data(), m_Status.size(), &textsize, DT_END_ELLIPSIS | DT_LEFT | DT_VCENTER | DT_NOPREFIX);
    SelectObject(dc, prev_font);

    EndPaint(hwnd, &ps);
}

/** General event handler. */
LRESULT
StatusPanel::handle(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
     case WM_PAINT:
        draw(hwnd, HDC(wparam));
        return 0;
     default:
        return Window::handle(hwnd, msg, wparam, lparam);
    }
}

/** Create window.
    \param x,y             Position
    \param nWidth,nHeight  Size
    \param hWndParent      Parent window */
HWND
StatusPanel::create(int x,                       // horizontal position of window
                    int y,                       // vertical position of window
                    int nWidth,                  // window width
                    int nHeight,                 // window height
                    HWND hWndParent)             // handle to parent or owner window
{
    m_Hwnd = Window::create(CLASS_NAME, "",
                            WS_CHILDWINDOW,
                            x, y, nWidth, nHeight, hWndParent, 0);
    return m_Hwnd;
}

/** Initialize StatusPanel object. Obtains all fonts and metrics, and
    registers the window class. */
void
StatusPanel::init()
{
    HDC desktop_dc = GetDC(GetDesktopWindow());

    int height = -MulDiv(10, GetDeviceCaps(desktop_dc, LOGPIXELSY), 72);

    m_BoldFont = CreateFont(height, 0, 0, 0,
                            FW_BOLD,
                            FALSE, FALSE, FALSE,
                            ANSI_CHARSET,
                            OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                            DEFAULT_QUALITY, DEFAULT_PITCH | FF_SWISS,
                            "Arial");
    m_NormalFont = CreateFont(height, 0, 0, 0,
                              FW_REGULAR,
                              FALSE, FALSE, FALSE,
                              ANSI_CHARSET,
                              OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                              DEFAULT_QUALITY, DEFAULT_PITCH | FF_SWISS,
                              "Arial");

    RECT size;
    size.left = size.top = 0;
    size.right = size.bottom = 99;
    HGDIOBJ prev = SelectObject(desktop_dc, m_BoldFont);
    DrawText(desktop_dc, "Tp", 2, &size, DT_CALCRECT);
    SelectObject(desktop_dc, prev);
    ReleaseDC(GetDesktopWindow(), desktop_dc);
    
    m_DefaultWidth  = 15 * size.right;
    m_DefaultHeight = 2 * size.bottom + 2*7 + 2;

    registerWindowClass(CLASS_NAME, 0);
}
