/**
  *  \file wingui.cpp
  *  \brief Windows GUI Main Entry Point
  *  \author Stefan Reuther
  *
  *  This module provides the main menu window and the logic to
  *  select game/race/action.
  */

#include <fstream>
#include <cstdio>
#include <cstring>
#include "api.h"
#include "wingui.h"
#include "nls.h"
#include "filenames.h"
#include "enumerator.h"
#include "game.h"
#include "loaderstate.h"
#include "action.h"
#include "race.h"
#include "program.h"
#include "util.h"
#include "netconfig.h"
#include "execute.h"
#include "workerthread.h"
#include "request.h"
#include "completionhandler.h"
#include "turnuploadrequest.h"
#include "statuspanel.h"
#include "evaluator.h"
#include "multirequest.h"

using namespace std;

enum { TIMER_NETWORK_POLL = 1234,
       TIMER_STATUS_PANEL = 2345 };

/** Constructor. Builds a main window editing \c data. */
MainWindow::MainWindow(Data& data)
    : m_Hwnd(0),
      m_Data(data),
      m_Init(data),
      m_FileMenu(0), m_GameMenu(0), m_RaceMenu(0), m_ActionMenu(0), m_ProgramMenu(0),
      m_HelpMenu(0), m_FullMenu(0),
      m_Terminated(false),
      m_nPanels(0),
      m_PanelYAdjust(0),
      m_PanelTimer(0),
      m_NetworkPanel(m_Exec)
{
    m_BigFont = CreateFont(24, 0, 0, 0, FW_BOLD, TRUE /*italic*/, FALSE, FALSE, DEFAULT_CHARSET,
                           OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FF_SWISS,
                           "Arial");
    m_SmallFont = CreateFont(16, 0, 0, 0, FW_BOLD, FALSE /*italic*/, FALSE, FALSE, DEFAULT_CHARSET,
                             OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FF_SWISS,
                             "Arial");
}

bool
MainWindow::askWinplan()
{
    int result = MessageBox(m_Hwnd,
                            (_("I could not assign game directories to ")
                             + itoa(m_Init.getGamesWithMissingDirs())
                             + _(" games, because Winplan has just room for 8 slots. "
                                 "You can let me assign a game directory outside the "
                                 "Winplan directory, which will then be accessible to "
                                 "programs other than Winplan (choose Yes). Alternatively, "
                                 "delete some games and retry later (choose No, then use "
                                 "File->Assign... to assign game directories).\r\n"
                                 "Assign directory outside Winplan tree?")).c_str(),
                            _("Game Setup"),
                            MB_ICONSTOP | MB_YESNO);
    return result == IDYES;
}

/** Create the main window. */
HWND
MainWindow::create()
{
    m_Hwnd = Window::create("MyWindowClass", "VGAP Integrator", WS_OVERLAPPEDWINDOW,
                            CW_USEDEFAULT, CW_USEDEFAULT, 400, 250,
                            0, getMenu());
    m_NetworkPanel.create(0, 0, 100, 100, m_Hwnd);
    recomputePanelSizes();
    return m_Hwnd;
}

static void
drawText(HDC dc, int x, int& y, std::string text)
{
    TEXTMETRIC tm;
    GetTextMetrics(dc, &tm);

    RECT r;
    r.left = x;
    r.right = x + 10000; // FIXME
    r.top = y;
    r.bottom = y + tm.tmHeight;
    DrawText(dc, text.data(), text.size(), &r, DT_LEFT | DT_NOPREFIX);

    y = r.bottom + 1;
}

LRESULT
MainWindow::handle(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
     case WM_CLOSE:
        m_Terminated = true;
        return 0;
     case WM_COMMAND:
        switch (UINT cmd = LOWORD(wparam)) {
         case ID_FILE_CLOSE:
            m_Terminated = true;
            return 0;
         case ID_FILE_LOAD:
            loadFile();
            return 0;
         case ID_FILE_ASSIGN_DIRS:
            m_Init.assignGameDirectories(false);
            postInit();
            return 0;
         case ID_FILE_NETCONFIG:
            doNetConfig(hwnd);
            return 0;
         case ID_HELP_ABOUT:
            {
                std::string s = PROGRAM_NAME;
                s += " ";
                s += PROGRAM_VERSION;
                s += "\r\n\r\n";
                s += _("Compiled ");
                s += __DATE__;
                s += " ";
                s += __TIME__;
                s += "\r\n\r\n";
                s += _("Settings are stored in ");
                s += getSettingsFileName();
                MessageBox(hwnd, s.c_str(), _("About"), MB_OK);
            }
            return 0;
         default:
            if (cmd >= ID_GAME && cmd < ID_GAME + m_Games.size()) {
                selectGame(m_Games[cmd - ID_GAME]);
                // FIXME: check game status
                if (m_Selection.getSelectedGame()->getLocalDirectory().size() == 0) {
                    MessageBox(hwnd, _("This game does not yet have a local directory assigned. "
                                       "Most functions do not work without one. "
                                       "Use File->Assign... to assign it a directory."),
                               _("Game Directory"),
                               MB_ICONSTOP | MB_OK);
                }
            } else if (cmd >= ID_RACE && cmd < ID_RACE + m_Races.size()) {
                selectRace(m_Races[cmd - ID_RACE]);
            } else if (cmd >= ID_ACTION && cmd < ID_ACTION + m_Actions.size()) {
                runAction(m_Actions[cmd - ID_ACTION]);
            } else if (cmd >= ID_PROGRAM && cmd < ID_PROGRAM + m_ProgramActions.size()) {
                runAction(m_ProgramActions[cmd - ID_PROGRAM]);
            }
            return 0;
        }
     case WM_PAINT: {
        PAINTSTRUCT ps;
        HDC dc = BeginPaint(hwnd, &ps);
        HGDIOBJ oldfont = SelectObject(dc, m_BigFont);
        int y = 30;
        int x = 30;
        if (Game* g = m_Selection.getSelectedGame())
            drawText(dc, x, y, g->getName());
        else
            drawText(dc, x, y, _("No game selected."));

        y += 10;
        SelectObject(dc, m_SmallFont);
        if (Race* r = m_Selection.getSelectedRace()) {
            drawText(dc, x, y, _("Player ") + itoa(r->getRaceNumber()));
        } else {
            drawText(dc, x, y, _("No player selected."));
        }
        if (Game* g = m_Selection.getSelectedGame()) {
            if (g->getLocalDirectory().size())
                drawText(dc, x, y, _("Game data in ") + g->getLocalDirectory());
            else
                drawText(dc, x, y, _("No game directory assigned."));
        }
        SelectObject(dc, oldfont);
        EndPaint(hwnd, &ps);
        return 0;
     }
     case WM_SIZE:
        if (wparam != SIZE_MINIMIZED)
            recomputePanelSizes();
        return 0;
     case WM_TIMER:
        if (wparam == TIMER_NETWORK_POLL) {
            // Stop timer to avoid recursion
            KillTimer(hwnd, TIMER_NETWORK_POLL);

            doRequestGC();

            // Restart timer
            SetTimer(hwnd, TIMER_NETWORK_POLL, 500, 0);
        } else if (wparam == TIMER_STATUS_PANEL) {
            KillTimer(hwnd, TIMER_STATUS_PANEL);
            doPanelTimer();
        }
        return 0;
     default:
        return Window::handle(hwnd, msg, wparam, lparam);
    }
}

/** Make sure \c menu is an empty menu.
    \param menu [in/out] Menu handle. Can be null.
    \post menu is an empty menu */
void
MainWindow::clearMenu(HMENU& menu)
{
    if (menu == 0) {
        menu = CreateMenu();
    } else {
        // TODO: prfen...
        do {
            HMENU m = GetSubMenu(menu, 0);
            if (m != 0)
                DestroyMenu(m);
        } while (RemoveMenu(menu, 0, MF_BYPOSITION));
    }
}

/** Build the File menu. */
void
MainWindow::buildFileMenu()
{
    clearMenu(m_FileMenu);
    AppendMenu(m_FileMenu, MF_STRING, ID_FILE_LOAD, _("&Load..."));
    AppendMenu(m_FileMenu, MF_STRING, ID_FILE_ASSIGN_DIRS, _("&Assign missing game directories"));
    AppendMenu(m_FileMenu, MF_SEPARATOR, 0, 0);
    AppendMenu(m_FileMenu, MF_STRING, ID_FILE_NETCONFIG, _("&Network configuration..."));
    AppendMenu(m_FileMenu, MF_SEPARATOR, 0, 0);
    AppendMenu(m_FileMenu, MF_STRING, ID_FILE_CLOSE, _("&Exit"));
}

/** Build the Game menu. */
void
MainWindow::buildGameMenu()
{
    clearMenu(m_GameMenu);

    m_Games.clear();
    VectorValueEnumerator<Game> e(m_Games);
    m_Data.enumerateGames(e);

    // HACK
    if (!m_Selection.getSelectedGame() && m_Games.size())
        m_Selection.selectGame(m_Games.front());

    for (unsigned int i = 0; i < m_Games.size(); ++i) {
        int flags;
        if (m_Selection.getSelectedGame() == m_Games[i])
            flags = MF_STRING | MF_CHECKED;
        else
            flags = MF_STRING;
        AppendMenu(m_GameMenu, flags, ID_GAME + i, m_Games[i]->getName().c_str());
    }

    if (GetMenuItemCount(m_GameMenu) == 0)
        AppendMenu(m_GameMenu, MF_STRING | MF_GRAYED, ID_GAME, "(empty)");
}

/** Build the Race menu. */
void
MainWindow::buildRaceMenu()
{
    clearMenu(m_RaceMenu);

    m_Races.clear();
    VectorValueEnumerator<Race> e(m_Races);
    if (Game* g = m_Selection.getSelectedGame())
        g->enumerateRaces(e);

    for (unsigned int i = 0; i < m_Races.size(); ++i) {
        int flags;
        if (m_Selection.getSelectedRace() == m_Races[i])
            flags = MF_STRING | MF_CHECKED;
        else
            flags = MF_STRING;

        char tmp[200];
        std::sprintf(tmp, _("Race %d"), m_Races[i]->getRaceNumber());

        AppendMenu(m_RaceMenu, flags, ID_RACE + i, tmp);
    }

    if (GetMenuItemCount(m_RaceMenu) == 0)
        AppendMenu(m_RaceMenu, MF_STRING | MF_GRAYED, ID_RACE, "(empty)");
}

/** Build the Action menu. */
void
MainWindow::buildActionMenu()
{
    clearMenu(m_ActionMenu);

    m_Actions.clear();
    VectorValueEnumerator<Action> a(m_Actions);
    // FIXME:
    if (Game* g = m_Selection.getSelectedGame())
        m_Data.enumerateSimpleActions(*g, a);

    for (unsigned int i = 0; i < m_Actions.size(); ++i) {
        int flags;
        m_Selection.selectAction(m_Actions[i]);
        if (m_Selection.expandString(m_Actions[i]->getCommandLine()).size())
            flags = MF_STRING;
        else
            flags = MF_STRING | MF_GRAYED;

        AppendMenu(m_ActionMenu, flags, ID_ACTION + i,
                   (m_Actions[i]->getDescription() + _(" (using ") + m_Actions[i]->getProgram().getName() + ")").c_str());
    }

    if (GetMenuItemCount(m_ActionMenu) == 0)
        AppendMenu(m_ActionMenu, MF_STRING | MF_GRAYED, ID_ACTION, "(empty)");
}

/** Build the Programs menu. */
void
MainWindow::buildProgramsMenu()
{
    clearMenu(m_ProgramMenu);

    std::vector<Program*> progs;
    VectorValueEnumerator<Program> a(progs);
    m_Data.enumeratePrograms(a);

    m_ProgramActions.clear();
    VectorValueEnumerator<Action> aa(m_ProgramActions);

    for (unsigned int i = 0; i < progs.size(); ++i) {
        // enumerate this program's actions, and store in array
        unsigned int org = m_ProgramActions.size();
        progs[i]->enumerateActions(aa);
        unsigned int count = m_ProgramActions.size() - org;

        HMENU the_menu;
        std::string pfx, sfx;
        if (count <= 1) {
            the_menu = m_ProgramMenu;
            pfx = progs[i]->getName() + " (";
            sfx = ")";
        } else {
            the_menu = CreateMenu();
            AppendMenu(m_ProgramMenu, MF_POPUP, UINT(the_menu), progs[i]->getName().c_str());
        }

        while (org < m_ProgramActions.size()) {
            int flags;
            m_Selection.selectAction(m_ProgramActions[org]);
            if (m_Selection.expandString(m_ProgramActions[org]->getCommandLine()).size())
                flags = MF_STRING;
            else
                flags = MF_STRING | MF_GRAYED;

            AppendMenu(the_menu, flags, ID_PROGRAM + org,
                       (pfx + m_ProgramActions[org]->getDescription() + sfx).c_str());
            ++org;
        }
    }

    if (GetMenuItemCount(m_ProgramMenu) == 0)
        AppendMenu(m_ProgramMenu, MF_STRING | MF_GRAYED, ID_PROGRAM, "(empty)");
}

/** Build the Help menu. */
void
MainWindow::buildHelpMenu()
{
    clearMenu(m_HelpMenu);
    AppendMenu(m_HelpMenu, MF_STRING, ID_HELP_ABOUT, "&About...");
}

/** Rebuild all variable menus after an unknown change of the current
    selection. */
void
MainWindow::rebuildAllMenus()
{
    buildGameMenu();
    buildRaceMenu();
    buildActionMenu();
    buildProgramsMenu();
}

/** Select a new game. */
void
MainWindow::selectGame(Game* g)
{
    m_Selection.selectGame(g);
    buildGameMenu();
    buildRaceMenu();
    buildActionMenu();
    buildProgramsMenu();
    InvalidateRect(m_Hwnd, 0, TRUE);
}

/** Select a new race. */
void
MainWindow::selectRace(Race* r)
{
    m_Selection.selectRace(r);
    buildRaceMenu();
    buildActionMenu();
    buildProgramsMenu();
    InvalidateRect(m_Hwnd, 0, TRUE);
}

/** Run an action. */
void
MainWindow::runAction(Action* a)
{
    m_Selection.selectAction(a);
    if (m_Exec.isPendingAction(m_Selection)) {
        MessageBox(m_Hwnd, "This action is still active.", a->getName().c_str(), MB_OK);
    } else {
        m_Exec.runAction(m_Hwnd, m_Selection);
        m_NetworkPanel.update();
    }
}

bool
MainWindow::askYesNo(const char* format, int arg)
{
    char tmp[300];
    std::sprintf(tmp, format, arg);
    return MessageBox(m_Hwnd, tmp, "Load Definition File", MB_YESNO | MB_ICONQUESTION) == IDYES;
}

void
MainWindow::loadFile(std::string name)
{
    std::ifstream in(name.c_str());
    if (!in) {
        MessageBox(m_Hwnd, (_("Unable to open file '") + name + "'").c_str(),
                   _("Load Definition File"), MB_ICONSTOP | MB_OK);
    } else {
        Transaction ta;
        WriteOperationEnumerator<Program> progwrite(ta);
        WriteOperationEnumerator<Game> gamewrite(ta);
        m_Data.enumeratePrograms(progwrite);
        m_Data.enumerateGames(gamewrite);
        if (!ta.tryAcquireLocks()) {
            MessageBox(m_Hwnd, _("Cannot load a definition file while background operations"
                                 " (network) are being performed. Please wait until the"
                                 " network operations finish and try again."),
                       _("Load Definition File"), MB_ICONSTOP | MB_OK);
            return;
        }
        try {
            Data tmp;
            LoaderState st;
            st.setBaseDirectory(getDirectoryName(name));
            tmp.read(in, st);

            if (st.getNumItems(st.Programs) == 0 && st.getNumItems(st.Games) == 0) {
                MessageBox(m_Hwnd, _("File does not contain any usable information."),
                           _("Load Definition File"), MB_ICONSTOP | MB_OK);
            } else {
                st.setAccept(st.Globals,  false);
                st.setAccept(st.Programs, st.getNumItems(st.Programs) != 0 &&
                             askYesNo(_("Accept %d program definitions?"), st.getNumItems(st.Programs)));
                st.setAccept(st.Games,    st.getNumItems(st.Games)    != 0 &&
                             askYesNo(_("Accept %d game definitions?"), st.getNumItems(st.Games)));
                if (st.isAcceptAny()) {
                    in.clear(); // weil wir am Ende waren
                    in.seekg(0);
                    m_Data.read(in, st);

                    if (st.isAccept(st.Games)) {
                        // FIXME: make this smarter
                        preInit();
                        postInit();
                    }

                    rebuildAllMenus();
                }
            }
        }
        catch (std::exception& e) {
            MessageBox(m_Hwnd,
                       (std::string(_("Load aborted due to error:\r\n")) + e.what()).c_str(),
                       _("Load Definition File"),
                       MB_ICONSTOP + MB_OK);
        }
    }
}

void
MainWindow::loadFile()
{
    char buffer[300];
    OPENFILENAME ofn;
    memset(&ofn, 0, sizeof(ofn));
    buffer[0] = 0;
    ofn.lStructSize          = sizeof(ofn);
    ofn.hwndOwner            = m_Hwnd;
    ofn.lpstrFilter          = "VGA Planets Game Definition Files (*.vgd)\0*.vgd\0"
                               "VGA Planets Program Definition Files (*.vpd)\0*.vpd\0"
                               "All Files\0*.*\0";
    // ofn.lpstrCustomFilter = 0;
    // ofn.nMaxCustFilter    = 0;
    // ofn.nFilterIndex      = 0;
    ofn.lpstrFile            = buffer;
    ofn.nMaxFile             = sizeof(buffer);
    // ofn.lpstrFileTitle    = 0;
    // ofn.nMaxFileTitle     = 0;
    // ofn.lpstrInitialDir   = 0;
    ofn.lpstrTitle           = _("Load Definition File");
    ofn.Flags                = OFN_FILEMUSTEXIST;
    // ofn.nFileOffset       = 0;
    // ofn.nFileExtension    = 0;
    // ofn.lpstrDefExt       = 0;

    if (GetOpenFileName(&ofn))
        loadFile(buffer);
}

/** Clean up finished requests. */
void
MainWindow::doRequestGC()
{
    bool did = false;
    while (Request* r = m_Exec.popFinishedRequest()) {
        // This 'wait' will not actually wait
        network_thread.getFinishSemaphore().wait();
        // Evaluate the request
        if (CompletionHandler* h = r->getCompletionHandler())
            h->handleCompletion();

        // Log it
        std::string title, status;
        COLORREF color;
        title = r->getName();

        Evaluator e;
        e.evaluateRequest(r);
        status = e.getDescription();

        static const COLORREF colors[] = { RGB(0, 128, 0), RGB(0, 128, 255), RGB(128, 0, 0), RGB(128, 0, 0) };
        color = colors[e.getStatus()];

        pushStatusPanel(new StatusPanel(title, status, color));
        did = true;
        
        delete r;
    }
    if (did)
        m_NetworkPanel.update();
}

/** Add new status panel. Automatically discards the oldest panel of
    needed to store the new one.

    \bug If panels arrive faster than we move them up, m_PanelYAdjust
    increases really fast, and no panel is visible for a certain
    time. We should probably cheat a little if this happens. */
void
MainWindow::pushStatusPanel(StatusPanel* p)
{
    if (m_nPanels >= NUM_PANELS) {
        // remove oldest panel
        delete m_Panels[--m_nPanels];
    }

    // insert new panel
    for (int i = m_nPanels; i > 0; --i)
        m_Panels[i] = m_Panels[i-1];
    m_Panels[0] = p;
    p->create(0, 0, 100, 100, m_Hwnd);
    ++m_nPanels;

    m_PanelYAdjust += StatusPanel::getDefaultHeight();

    recomputePanelSizes();

    SetTimer(m_Hwnd, TIMER_STATUS_PANEL, 50, 0);
    m_PanelTimer = 50;
}

/** Reposition all the panels.

    \todo the NetworkPanel should be below the StatusPanels, and
    always be hidden by the latter. Unfortunately, I cannot get
    that to work reliably. Fix this later. */
void
MainWindow::recomputePanelSizes()
{
    RECT mysize;
    GetClientRect(m_Hwnd, &mysize);

    for (int i = 0; i < m_nPanels; ++i) {
        SetWindowPos(m_Panels[i]->getWindowHandle(),
                     0,
                     mysize.right - StatusPanel::getDefaultWidth(),
                     mysize.bottom - (i+1) * StatusPanel::getDefaultHeight() + m_PanelYAdjust,
                     StatusPanel::getDefaultWidth(),
                     StatusPanel::getDefaultHeight(),
                     SWP_NOOWNERZORDER | SWP_NOACTIVATE | SWP_SHOWWINDOW | SWP_NOCOPYBITS);
    }

    SetWindowPos(m_NetworkPanel.getWindowHandle(),
                 0,
                 mysize.right - NetworkPanel::getDefaultWidth(),
                 mysize.bottom - NetworkPanel::getDefaultHeight(),
                 NetworkPanel::getDefaultWidth(),
                 NetworkPanel::getDefaultHeight(),
                 SWP_NOOWNERZORDER | SWP_NOACTIVATE | SWP_SHOWWINDOW | SWP_NOCOPYBITS);
}

/** Panel timer. Called after \c m_PanelTimer ms have elapsed. This
    moves the panels up if required. */
void
MainWindow::doPanelTimer()
{
    // increment age of all panels
    for (int i = 0; i < m_nPanels; ++i)
        m_Panels[i]->addAge(m_PanelTimer);

    // remove old panels
    while (m_nPanels != 0 && m_Panels[m_nPanels - 1]->isOld())
        delete m_Panels[--m_nPanels];

    // move up?
    if (m_PanelYAdjust > 0) {
        if (m_PanelYAdjust > 2)
            m_PanelYAdjust -= 2;
        else
            m_PanelYAdjust = 0;
        recomputePanelSizes();
        SetTimer(m_Hwnd, TIMER_STATUS_PANEL, 50, 0);
        m_PanelTimer = 50;
    } else if (m_nPanels > 0) {
        // because arrival of a new panel resets the timer or
        // otherwise disturbs it, we cap the time-to-wait at
        // 1000. This way, at most 1000 ms are lost. Otherwise, a new
        // panel arriving after the current one was visible for 14
        // seconds will cancel these 14 seconds, leaving the old one
        // around for 30 seconds total.
        int n = m_Panels[m_nPanels - 1]->getTimeUntilOld();
        if (n > 1000)
            n = 1000;
        SetTimer(m_Hwnd, TIMER_STATUS_PANEL, n, 0);
        m_PanelTimer = n;
    }
}

/** Create the complete menu. */
HMENU
MainWindow::getMenu()
{
    if (m_FullMenu == 0) {
        buildFileMenu();
        buildGameMenu();
        buildRaceMenu();
        buildActionMenu();
        buildProgramsMenu();
        buildHelpMenu();

        m_FullMenu = CreateMenu();
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_FileMenu), _("&File"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_GameMenu), _("&Game"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_RaceMenu), _("&Race"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_ActionMenu), _("&Action"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_ProgramMenu), _("&Programs"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_HelpMenu), _("&Help"));
    }
    return m_FullMenu;
}

void
MainWindow::preInit()
{
    m_Init.assignGameDirectories(false);
}

void
MainWindow::postInit()
{
    int n = m_Init.getGamesWithMissingDirs();
    if (n != 0) {
        if (askWinplan()) {
            m_Init.assignGameDirectories(true);
        }
    }
    InvalidateRect(m_Hwnd, 0, TRUE);
    rebuildAllMenus();

    n = m_Init.getGamesWithMissingSpecs();
    if (n != 0) {
        int result = MessageBox(m_Hwnd,
                                (itoa(n) + _(" games have just been set up, but their data files "
                                             "have not yet been downloaded and installed. Do you "
                                             "want me to download these files now?")).c_str(),
                                _("Game Setup"),
                                MB_ICONSTOP | MB_YESNO);
        if (result == IDYES) {
            Selection tmp_sel;
            MultiRequest mr(_("Download specifications"));
            m_Init.downloadMissingSpecs(mr);
            if (mr.getNumRequests()) {
                m_Exec.pushRequest(tmp_sel, new MultiRequest(mr));
                m_NetworkPanel.update();
            }
        }
    }
}

static void
loadDefaults(Data& d)
{
    std::string name = getSettingsFileName();
    std::ifstream in(name.c_str());
    if (in) {
        LoaderState st;
        st.setBaseDirectory(getDirectoryName(name));
        d.read(in, st);
    }
    d.ensureDefaultBindings(false);
}

static void
saveDefaults(Data& d)
{
    std::string filename = getSettingsFileName();
    std::string bak = filename + ".bak";
    std::remove(bak.c_str());
    std::rename(filename.c_str(), bak.c_str());
    std::ofstream out(filename.c_str());
    if (out)
        d.write(out);
}

static const char pipe_name[] = "\\\\.\\pipe\\VGAPIntegratorPipe";
static const char mutex_name[] = "VGAPIntegratorMutex";

struct CommStruct {
    DWORD m_Command;
    DWORD m_PayloadSize;
};
enum {
    CommandRaise,
    CommandLoad
};

/** Check whether there is another instance. If there is one, tell it
    to do what user wants us to do, and return true; GUI shouldn't
    start up in that case. Otherwise, returns false. */
bool
checkOtherInstance(int argc, char** argv)
{
    HANDLE hpipe = INVALID_HANDLE_VALUE;

    HANDLE hmutex = CreateMutex(0, FALSE, mutex_name);
    if (hmutex == 0)
        goto fail;

    if (GetLastError() == ERROR_ALREADY_EXISTS) {
        // there is already an instance
        if (!WaitNamedPipe(pipe_name, 1000)) {
            MessageBox(0, _("Unable to communicate with running instance. It is possibly already "
                            "processing another request."),
                       _("VGAP Integrator"), MB_ICONSTOP | MB_OK);
            CloseHandle(hmutex);
            std::exit(1);
        }
        hpipe = CreateFile(pipe_name, GENERIC_READ | GENERIC_WRITE,
                           0, 0, OPEN_EXISTING, 0, 0);
        if (hpipe == INVALID_HANDLE_VALUE)
            goto fail;

        // Send request to other end
        CommStruct cs;
        DWORD written;
        if (argc < 2) {
            cs.m_Command = CommandRaise;
            cs.m_PayloadSize = 0;
            WriteFile(hpipe, &cs, sizeof(cs), &written, 0);
        } else {
            cs.m_Command = CommandLoad;
            cs.m_PayloadSize = strlen(argv[1]);
            WriteFile(hpipe, &cs, sizeof(cs), &written, 0);
            WriteFile(hpipe, argv[1], strlen(argv[1]), &written, 0);
        }

        // Wait for reply
        char tmp;
        ReadFile(hpipe, &tmp, 1, &written, 0);

        CloseHandle(hmutex);
        CloseHandle(hpipe);
        
        return true;
    } else {
        // does not exist, we are the master
        return false;
    }

fail:
    if (hmutex != 0)
        CloseHandle(hmutex);
    if (hpipe != INVALID_HANDLE_VALUE)
        CloseHandle(hpipe);
    MessageBox(0, _("Unable to initialize application."), _("VGAP Integrator"), MB_ICONSTOP | MB_OK);
    std::exit(1);
    return true;
}

int
RealMain(int argc, char** argv)
{
    if (checkOtherInstance(argc, argv))
        return 0;

    // Initialize widgets
    NetworkPanel::init();
    Window::registerWindowClass("MyWindowClass", 0);

    // Load data
    Data data;
    loadDefaults(data);

    // Create main window
    MainWindow mainwin(data);
    HWND wnd = mainwin.create();
    ShowWindow(wnd, SW_SHOW);
    mainwin.preInit();

    if (argc >= 2) {
        mainwin.loadFile(argv[1]);
    }

    mainwin.postInit();

    MSG msg;

    // Network stuff
    network_thread.start();
    SetTimer(wnd, TIMER_NETWORK_POLL, 500, 0);

    // Main loop
    HANDLE pipe = INVALID_HANDLE_VALUE;
    HANDLE evt = CreateEvent(0, TRUE, FALSE, 0);
    OVERLAPPED ovl;
    bool have_pipe = false;
    while (!mainwin.isTerminated()) {
        if (!have_pipe) {
            pipe = CreateNamedPipe(pipe_name, PIPE_ACCESS_DUPLEX | FILE_FLAG_OVERLAPPED,
                                   PIPE_WAIT | PIPE_TYPE_BYTE, PIPE_UNLIMITED_INSTANCES,
                                   4096, 4096, 1000, 0);
            ResetEvent(evt);
            have_pipe = true;
            if (pipe != INVALID_HANDLE_VALUE) {
                ovl.hEvent = evt;
                ConnectNamedPipe(pipe, &ovl);
            }
        }

        int result = MsgWaitForMultipleObjects(1, &evt, FALSE, INFINITE, QS_ALLINPUT | QS_ALLPOSTMESSAGE);
        if (result == WAIT_OBJECT_0) {
            CommStruct cs;
            DWORD read;
            if (ReadFile(pipe, &cs, sizeof(cs), &read, 0) && read == sizeof(cs)) {
                char payload[1024];
                DWORD payloadsize;
                if (cs.m_PayloadSize > 0 && cs.m_PayloadSize < sizeof(payload)) {
                    if (!ReadFile(pipe, payload, cs.m_PayloadSize, &payloadsize, 0))
                        payloadsize = 0;
                } else {
                    payloadsize = 0;
                }
                if (payloadsize != cs.m_PayloadSize)
                    MessageBox(0, "warning: invalid payload size", "comm error", MB_OK);

                if (cs.m_Command == CommandRaise) {
                    // raise window. XP only blinks the window's task
                    // bar item, though :-(
                    SetForegroundWindow(wnd);
                } else if (cs.m_Command == CommandLoad) {
                    // load file
                    mainwin.loadFile(std::string(payload, payloadsize));
                }

                // Send reply
                WriteFile(pipe, payload, 1, &payloadsize, 0);
            }
            CloseHandle(pipe);
            have_pipe = false;
        } else if (result == WAIT_OBJECT_0+1) {
            while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        }
    }

    saveDefaults(data);

    return 0;
}
